/*
nRF24L01 connections
 1 - GND
 2 - VCC 3.3V !!! Ideally 3.0v, definitely not 5V
 3 - CE to Arduino pin 9
 4 - CSN to Arduino pin 10
 5 - SCK to Arduino pin 13
 6 - MOSI to Arduino pin 11
 7 - MISO to Arduino pin 12
 8 - UNUSED
*/

#include <SPI.h>
#include <nRF24L01.h>
#include <RF24.h>
#include "printf.h"

unsigned int MS5611_calibrationData[6];

/*
// breadboard build
uint16_t calib_T1 = 28170;
int16_t  calib_T2 = 26049;
int16_t  calib_T3 = 50;
uint16_t calib_P1 = 37410;
int16_t  calib_P2 = -10890;
int16_t  calib_P3 = 3024;
int16_t  calib_P4 = 5231;
int16_t  calib_P5 = 1;
int16_t  calib_P6 = -7;
int16_t  calib_P7 = 15500;
int16_t  calib_P8 = -14600;
int16_t  calib_P9 = 6000;
*/

// PCB build
uint16_t calib_T1 = 27932;
int16_t  calib_T2 = 26185;
int16_t  calib_T3 = 50;
uint16_t calib_P1 = 37554;
int16_t  calib_P2 = -10840;
int16_t  calib_P3 = 3024;
int16_t  calib_P4 = 5498;
int16_t  calib_P5 = -30;
int16_t  calib_P6 = -7;
int16_t  calib_P7 = 15500;
int16_t  calib_P8 = -14600;
int16_t  calib_P9 = 6000;

const uint64_t pipeIn =  0xE8E8F0F0E1LL;

RF24 radio(9, 10);

enum _packetType {
  PT_GPS,
  PT_SENSORS
};

struct _sensors {
  byte type;
  int MPU6050_accelerometer[3];
  int MPU6050_temperature;
  uint32_t MS5611_D1;
  uint32_t MS5611_D2;
  int32_t BMP280_adc_T;
  int32_t BMP280_adc_P;
  byte csq;
  byte creg;
  unsigned int vbat;
};
  
struct _gps {
    byte type;
    unsigned char fixType;      // GNSSfix Type, range 0..5
    unsigned char numSV;        // Number of satellites used in Nav Solution
   
    long lon;                   // Longitude (deg)
    long lat;                   // Latitude (deg)
    long hMSL;                  // Height above mean sea level (mm)
   
    int velD;                   // NED down velocity (dm/s)
    int gSpeed;                 // Ground Speed (2-D) (dm/s)
    int heading;                // Heading of motion 2-D (deg 1e-4)
    
    int Si7021_humidity;
    int Si7021_temperature;
    int DS18B20_temperature1;
    int DS18B20_temperature2;
};

// The sizeof this struct should not exceed 32 bytes
union Packet {
  _gps gps;
  _sensors sensors;
};

Packet packet;
_gps gps;
_sensors sensors;

int32_t t_fine;

float readTemperature()
{
  int32_t var1, var2;

  int32_t adc_T = sensors.BMP280_adc_T;
  adc_T >>= 4;

  var1  = ((((adc_T>>3) - ((int32_t)calib_T1 <<1))) *
     ((int32_t)calib_T2)) >> 11;

  var2  = (((((adc_T>>4) - ((int32_t)calib_T1)) *
       ((adc_T>>4) - ((int32_t)calib_T1))) >> 12) *
     ((int32_t)calib_T3)) >> 14;

  t_fine = var1 + var2;

  float T  = (t_fine * 5 + 128) >> 8;
  return T/100;
}

float readPressure() {
  int64_t var1, var2, p;

  // Must be done first to get the t_fine variable set up
  readTemperature();

  int32_t adc_P = sensors.BMP280_adc_P;
  adc_P >>= 4;

  var1 = ((int64_t)t_fine) - 128000;
  var2 = var1 * var1 * (int64_t)calib_P6;
  var2 = var2 + ((var1*(int64_t)calib_P5)<<17);
  var2 = var2 + (((int64_t)calib_P4)<<35);
  var1 = ((var1 * var1 * (int64_t)calib_P3)>>8) +
    ((var1 * (int64_t)calib_P2)<<12);
  var1 = (((((int64_t)1)<<47)+var1))*((int64_t)calib_P1)>>33;

  if (var1 == 0) {
    return 0;  // avoid exception caused by division by zero
  }
  p = 1048576 - adc_P;
  p = (((p<<31) - var2)*3125) / var1;
  var1 = (((int64_t)calib_P9) * (p>>13) * (p>>13)) >> 25;
  var2 = (((int64_t)calib_P8) * p) >> 19;

  p = ((p + var1 + var2) >> 8) + (((int64_t)calib_P7)<<4);
  return (float)p/256;
}

/**************************************************/

void setup()
{
  Serial.begin(115200);
  printf_begin();

/*
//breadboard build
  MS5611_calibrationData[0] = 45844;
  MS5611_calibrationData[1] = 47749;
  MS5611_calibrationData[2] = 28148;
  MS5611_calibrationData[3] = 25102;
  MS5611_calibrationData[4] = 31205;
  MS5611_calibrationData[5] = 27597;
*/

// PCB build
  MS5611_calibrationData[0] = 54312;
  MS5611_calibrationData[1] = 56861;
  MS5611_calibrationData[2] = 33953;
  MS5611_calibrationData[3] = 32753;
  MS5611_calibrationData[4] = 32333;
  MS5611_calibrationData[5] = 28374;

 
  radio.begin();
  radio.setDataRate(RF24_250KBPS);
  radio.setAutoAck(false);
  radio.openReadingPipe(1,pipeIn);
  radio.startListening();

  //radio.printDetails();
}

/**************************************************/

void comma() {
  Serial.print(", ");
}

void logPackets() {

    Serial.print("#SV: ");      Serial.print(gps.numSV);
    Serial.print(" fixType: "); Serial.print(gps.fixType);
    Serial.print(" lat/lon: "); Serial.print(gps.lat); Serial.print(","); Serial.print(gps.lon);
    Serial.print(" gSpeed: ");  Serial.print(gps.gSpeed/1000.0f);
    Serial.print(" heading: "); Serial.println(gps.heading/100000.0f);
    
    Serial.print(gps.Si7021_humidity / 100.0f);
    comma();
    Serial.print(gps.Si7021_temperature / 100.0f);
    comma();
    Serial.print(gps.DS18B20_temperature1 / 100.0f);
    comma();
    Serial.print(gps.DS18B20_temperature2 / 100.0f);
    Serial.println();

    int32_t dT;
    int64_t OFF, SENS;
    long P, TEMP;

    dT = sensors.MS5611_D2 - ((long)MS5611_calibrationData[4] << 8);
    TEMP = 2000 + (((int64_t)dT * (int64_t)MS5611_calibrationData[5]) >> 23);
    OFF = ((unsigned long)MS5611_calibrationData[1] << 16) + (((int64_t)MS5611_calibrationData[3] * dT) >> 7);
    SENS = ((unsigned long)MS5611_calibrationData[0] << 15) + (((int64_t)MS5611_calibrationData[2] * dT) >> 8);
    P = ((sensors.MS5611_D1 * (SENS >> 21)) - OFF) >> 15;
   
    Serial.print(sensors.MPU6050_accelerometer[0] / 16384.0f);
    comma();
    Serial.print(sensors.MPU6050_accelerometer[1] / 16384.0f);
    comma();
    Serial.print(sensors.MPU6050_accelerometer[2] / 16384.0f);
    comma();
    Serial.print( (sensors.MPU6050_temperature/340.0f)+35 );
    comma();
    Serial.print(TEMP / 100.0f);
    comma();
    Serial.print(P / 100.0f);
    comma();
    Serial.print(readTemperature());
    comma();
    Serial.println(readPressure() / 100.0f);
    Serial.println();
}
/*
void logJSON() {
  Serial.print("{\"packets\":"); Serial.print(packetsRead);
  Serial.print(",\"hour\":"); Serial.print(gps.fixType);
  Serial.print(",\"min\":"); Serial.print(gps.numSV);
  //Serial.print(",\"lat\":"); Serial.print(gps.lat);
  Serial.print(",\"sec\":"); Serial.print(gps.lon);
  Serial.println("}");
}
*/

float mbarToMeters(float mbar) {
  float fraction = mbar / 1013.25f;
  float raised = pow( fraction, 0.190284 );
  return 145366.45 * ( 1 - raised );
}

unsigned long packetsRead = 0;
unsigned long lastRecvTime = 0;
unsigned long lastDisplayTime = 0;

void logJSON() {

  int32_t dT;
  int64_t OFF, SENS;
  long P, TEMP;

  dT = sensors.MS5611_D2 - ((long)MS5611_calibrationData[4] << 8);
  TEMP = 2000 + (((int64_t)dT * (int64_t)MS5611_calibrationData[5]) >> 23);
  OFF = ((unsigned long)MS5611_calibrationData[1] << 16) + (((int64_t)MS5611_calibrationData[3] * dT) >> 7);
  SENS = ((unsigned long)MS5611_calibrationData[0] << 15) + (((int64_t)MS5611_calibrationData[2] * dT) >> 8);
  P = ((sensors.MS5611_D1 * (SENS >> 21)) - OFF) >> 15;
  
  Serial.print("{\"dt\":"); Serial.print(millis() - lastRecvTime);
  Serial.print(",\"packets\":"); Serial.print(packetsRead);
  Serial.print(",\"fix\":"); Serial.print(gps.fixType);
  Serial.print(",\"sv\":"); Serial.print(gps.numSV);
  Serial.print(",\"lat\":"); Serial.print(gps.lat);
  Serial.print(",\"lon\":"); Serial.print(gps.lon);
  Serial.print(",\"hmsl\":"); Serial.print(gps.hMSL);
  Serial.print(",\"veld\":"); Serial.print(gps.velD);
  Serial.print(",\"gspeed\":"); Serial.print(gps.gSpeed);
  Serial.print(",\"heading\":"); Serial.print(gps.heading);

  Serial.print(",\"csq\":"); Serial.print(sensors.csq);
  Serial.print(",\"creg\":"); Serial.print(sensors.creg);
  
  Serial.print(",\"Si7021_h\":"); Serial.print(gps.Si7021_humidity / 100.0f);
  Serial.print(",\"Si7021_t\":"); Serial.print(gps.Si7021_temperature / 100.0f);
  Serial.print(",\"ds1\":"); Serial.print(gps.DS18B20_temperature1 / 100.0f);
  Serial.print(",\"ds2\":"); Serial.print(gps.DS18B20_temperature2 / 100.0f);

  Serial.print(",\"MPU6050_ax\":"); Serial.print(sensors.MPU6050_accelerometer[0] / 16384.0f);
  Serial.print(",\"MPU6050_ay\":"); Serial.print(sensors.MPU6050_accelerometer[1] / 16384.0f);
  Serial.print(",\"MPU6050_az\":"); Serial.print(sensors.MPU6050_accelerometer[2] / 16384.0f);
  Serial.print(",\"MPU6050_t\":"); Serial.print( (sensors.MPU6050_temperature / 340.0f) + 35 );
  Serial.print(",\"MS5611_t\":"); Serial.print(TEMP / 100.0f);
  Serial.print(",\"MS5611_p\":"); Serial.print(P / 100.0f);
  Serial.print(",\"BMP280_t\":"); Serial.print(readTemperature());
  Serial.print(",\"BMP280_p\":"); Serial.print(readPressure() / 100.0f);

  float r1 = 3280;
  float r2 = 9991;
  
  Serial.print(",\"vbat\":"); Serial.print(((sensors.vbat/1023.0f)*3.3) * ((r1+r2)/r1) );
  
  Serial.println("}");

}

void recvData()
{
  while ( radio.available() ) {        
    radio.read(&packet, sizeof(Packet));
    packetsRead++;
    if ( packet.gps.type == PT_GPS ) 
      gps = packet.gps;
    else
      sensors = packet.sensors;
    lastRecvTime = millis();
  }    
}

void loop()
{
  recvData();
  
  unsigned long now = millis();
  if ( now - lastDisplayTime > 200 ) {
    //if ( packetsRead > 0 ) {
      logJSON();
      packetsRead = 0;
    //}
    lastDisplayTime = millis();
  }
  
}

